#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import io
import logging
import os
import os.path as osp

from .common import Metadata
from .common import dist_info_name
from .common import write_entry_points
from .config import read_pyproject_file
from .sdist import make_sdist
from .wheel import _write_WHEEL_file
from .wheel import make_wheel

log = logging.getLogger(__name__)

"""PEP-517 compliant buildsystem API"""


# We have no extra building reuirements and to build sdist, wheel or editable
# Requirements to build an editable are the same as for a wheel
def get_requires_for_build_wheel(config_settings=None):
    return []


get_requires_for_build_sdist = get_requires_for_build_wheel
get_requires_for_build_editable = get_requires_for_build_wheel


def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None):
    """Creates {metadata_directory}/foo-1.2.dist-info"""
    config_settings = config_settings or {}
    pyproject_file = config_settings.get("--pyproject")
    wheel_tag = config_settings.get("--wheel-tag") or "py3-any-none"

    project_info = read_pyproject_file(pyproject_file)
    metadata = Metadata(project_info.metadata)

    dist_info_dir = osp.join(metadata_directory, dist_info_name(metadata.name, metadata.version))
    os.mkdir(dist_info_dir)

    if project_info.entrypoints:
        with io.open(osp.join(dist_info_dir, "entry_points.txt"), "w", encoding="utf-8") as f:
            write_entry_points(project_info.entrypoints, f)

    with io.open(osp.join(dist_info_dir, "WHEEL"), "w", encoding="utf-8") as f:
        _write_WHEEL_file(f, wheel_tag=wheel_tag)

    with io.open(osp.join(dist_info_dir, "METADATA"), "w", encoding="utf-8") as f:
        metadata.write_metadata_file(f)

    return osp.basename(dist_info_dir)


# Metadata for editable are the same as for a wheel
prepare_metadata_for_build_editable = prepare_metadata_for_build_wheel


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    config_settings = config_settings or {}
    pyproject_file = config_settings.get("--pyproject")
    wheel_tag = config_settings.get("--wheel-tag")

    wheel_file = make_wheel(
        pyproject_file=pyproject_file,
        output_dir=wheel_directory,
        wheel_tag=wheel_tag,
    )
    return wheel_file.name


def build_editable(wheel_directory, config_settings=None, metadata_directory=None):
    config_settings = config_settings or {}
    pyproject_file = config_settings.get("--pyproject")
    wheel_tag = config_settings.get("--wheel-tag")

    wheel_file = make_wheel(
        pyproject_file=pyproject_file,
        output_dir=wheel_directory,
        wheel_tag=wheel_tag,
        editable=True,
    )
    return wheel_file.name


def build_sdist(sdist_directory, config_settings=None):
    config_settings = config_settings or {}
    pyproject_file = config_settings.get("--pyproject")

    sdist_file = make_sdist(
        pyproject_file=pyproject_file,
        output_dir=sdist_directory,
    )
    return sdist_file.name



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import os
import os.path as osp
import zipfile
from contextlib import contextmanager

from testpath import assert_isdir
from testpath import assert_isfile
from testpath.tempdir import TemporaryDirectory

from flot import buildapi

data_dir = osp.join(osp.dirname(__file__), "data")


@contextmanager
def cwd(directory):
    prev = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(prev)


def test_build_wheel():
    with TemporaryDirectory() as td, cwd(osp.join(data_dir, "pep517")):
        filename = buildapi.build_wheel(td)
        assert str(filename).endswith(".whl"), filename
        assert_isfile(osp.join(td, filename))
        assert zipfile.is_zipfile(osp.join(td, filename))
        with zipfile.ZipFile(osp.join(td, filename)) as zipped:
            assert "module1.py" in zipped.namelist()
            assert "module1.pth" not in zipped.namelist()


def test_build_wheel_pep621():
    with TemporaryDirectory() as td, cwd(osp.join(data_dir, "pep621")):
        filename = buildapi.build_wheel(td)
        assert str(filename).endswith(".whl"), filename
        assert_isfile(osp.join(td, filename))
        assert zipfile.is_zipfile(osp.join(td, filename))


def test_prepare_metadata_for_build_wheel():
    with TemporaryDirectory() as td, cwd(osp.join(data_dir, "pep517")):
        dirname = buildapi.prepare_metadata_for_build_wheel(td)
        assert dirname.endswith(".dist-info"), dirname
        assert_isdir(osp.join(td, dirname))
        assert_isfile(osp.join(td, dirname, "METADATA"))



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

from pathlib import Path

import pytest

from flot.common import FileSelector
from flot.common import InvalidVersion
from flot.common import NoVersionError
from flot.common import check_version
from flot.common import normalize_file_permissions

data_dir = Path(__file__).parent / "data"


def test_version_raise():
    with pytest.raises(InvalidVersion):
        check_version("a.1.0.beta0")

    with pytest.raises(InvalidVersion):
        check_version("3!")

    with pytest.raises(InvalidVersion):
        check_version((1, 2))

    with pytest.raises(NoVersionError):
        check_version(None)

    assert check_version("4.1.0beta1") == "4.1.0b1"
    assert check_version("v1.2") == "1.2"


def test_normalize_file_permissions():
    assert normalize_file_permissions(0o100664) == 0o100644  # regular file
    assert normalize_file_permissions(0o40775) == 0o40755  # directory


def test_FileSelector_ignores_pycache_and_pyc(copy_test_data):
    test_dir = Path(copy_test_data("module3"))

    pyc1 = test_dir / "python1.pyc"
    with pyc1.open("w") as f:
        f.write("")

    pycache = test_dir / "__pycache__"
    pycache.mkdir()

    pyc2 = pycache / "python2.pyc"
    with pyc2.open("w") as f:
        f.write("")

    expected = ["LICENSE", "pyproject.toml", "src/module3.py"]
    fs = FileSelector(
        base_dir=test_dir,
        includes=["**/*"],
    )
    results = [str(frel) for _fabs, frel in fs.files]
    assert results == expected


def test_FileSelector_honors_includes(get_test_data):
    base_dir = get_test_data("module3")
    fs = FileSelector(
        base_dir=base_dir,
        includes=["pyproject.toml"],
    )
    results = [str(frel) for _fabs, frel in fs.files]
    assert results == ["pyproject.toml"]


def test_FileSelector_honors_includes_and_excludes(get_test_data):
    base_dir = get_test_data("module4")
    fs = FileSelector(
        base_dir=base_dir,
        includes=["pyproject.toml", "src/**/*"],
    )
    expected = ["pyproject.toml", "src/module3.py", "src/module4.py"]
    results = [str(frel) for _fabs, frel in fs.files]
    assert results == expected

    fs = FileSelector(
        base_dir=base_dir,
        includes=["pyproject.toml", "src/**/*"],
        excludes=["src/module3.py"],
    )
    results = [str(frel) for _fabs, frel in fs.files]
    expected = ["pyproject.toml", "src/module4.py"]
    assert results == expected


def test_FileSelector_honors_includes_excludes_with_nested_file_tree(get_test_data):
    base_dir = get_test_data("module5")

    fs = FileSelector(
        base_dir=base_dir,
        includes=["pyproject.toml", "src/**/*"],
    )
    results = [str(frel) for _fabs, frel in fs.files]
    expected = [
        "pyproject.toml",
        "src/deep/nested/foo.py",
        "src/module3.py",
        "src/module4.py",
    ]
    assert results == expected

    fs = FileSelector(
        base_dir=base_dir,
        includes=["pyproject.toml", "src/**/*"],
        excludes=["src/module3.py", "src/deep/**/*"],
    )

    results = [str(frel) for _fabs, frel in fs.files]
    expected = ["pyproject.toml", "src/module4.py"]
    assert results == expected



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import configparser
import os
import stat
import tempfile
import zipfile
from pathlib import Path
from unittest import skipIf
from zipfile import ZipFile

from testpath import assert_isdir
from testpath import assert_isfile

from flot.wheel import WheelBuilder
from flot.wheel import make_wheel

data_dir = Path(__file__).parent / "data"


def test_WheelBuilder_select_files():
    pyproject_file = data_dir / "inclusion" / "pyproject.toml"
    base_dir = pyproject_file.parent
    wb = WheelBuilder.from_pyproject_file(pyproject_file)
    files = [(str(a.relative_to(base_dir)), str(b)) for a, b in wb.selected_files.files]
    assert files == [("module1.py", "module1.py")]


def test_inclusion_dir(tmp_path):
    # Smoketest for https://github.com/pypa/flit/issues/399
    wheel = make_wheel(
        pyproject_file=data_dir / "inclusion" / "pyproject.toml", output_dir=tmp_path
    )
    assert_isfile(wheel)


def test_source_date_epoch(tmp_path, monkeypatch):
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "1633007882")
    wheel = make_wheel(pyproject_file=data_dir / "pep621" / "pyproject.toml", output_dir=tmp_path)
    assert_isfile(wheel)
    # Minimum value for zip timestamps is 1980-1-1
    with ZipFile(wheel, "r") as zf:
        assert zf.getinfo("module1a.py").date_time == (2022, 2, 2, 2, 2, 2)


def test_zero_timestamp(tmp_path, monkeypatch):
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "0")
    wheel = make_wheel(pyproject_file=data_dir / "pep621" / "pyproject.toml", output_dir=tmp_path)
    assert_isfile(wheel)
    # Minimum value for zip timestamps is 1980-1-1
    with ZipFile(wheel, "r") as zf:
        assert zf.getinfo("module1a.py").date_time == (2022, 2, 2, 2, 2, 2)


def test_data_dir(tmp_path):
    wheel = make_wheel(
        pyproject_file=data_dir / "with_data_dir" / "pyproject.toml",
        output_dir=tmp_path,
    )
    assert_isfile(wheel)
    with ZipFile(wheel, "r") as zf:
        names = zf.namelist()
        assert "data/share/man/man1/foo.1" in names
        assert "module1-1.2.dist-info/LICENSE" in names
        assert "module1-1.2.dist-info/README.rst" not in names


def unpack(path):
    z = zipfile.ZipFile(str(path))
    t = tempfile.TemporaryDirectory()
    z.extractall(t.name)
    return t


def test_wheel_module(copy_test_data):
    td = copy_test_data("module1_toml")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    expected = td / "module1-1.0-py3-none-any.whl"
    assert_isfile(wf)
    assert str(wf) == str(expected)

    with unpack(wf) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        assert files == [
            "module1-1.0.dist-info/METADATA",
            "module1-1.0.dist-info/RECORD",
            "module1-1.0.dist-info/WHEEL",
            "module1.py",
        ]


def test_editable_wheel_module(copy_test_data):
    td = copy_test_data("module1_toml")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td, editable=True)
    whl_file = td / "module1-1.0-py3-none-any.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        pth_path = Path(unpacked, "module1.pth")
        assert_isfile(pth_path)
        assert pth_path.read_text().endswith("module1.py\n")
        assert_isdir(Path(unpacked, "module1-1.0.dist-info"))


def test_editable_wheel_has_absolute_pth(copy_test_data):
    td = copy_test_data("module1_toml")
    oldcwd = os.getcwd()
    os.chdir(td)
    try:
        wf = make_wheel(pyproject_file=Path("pyproject.toml"), output_dir=td, editable=True)
        whl_file = Path("module1-1.0-py3-none-any.whl").absolute()
        assert_isfile(wf)
        assert str(wf) == str(whl_file)
        with unpack(whl_file) as unpacked:
            path_file = Path(unpacked, "module1.pth")
            check_paths(path_file, td)
            assert_isdir(Path(unpacked, "module1-1.0.dist-info"))
    finally:
        os.chdir(oldcwd)


def check_paths(path_file, test_dir):
    assert_isfile(path_file)
    for pth in path_file.read_text().splitlines(False):
        p = Path(pth)
        assert p.is_absolute()
        assert p.exists()
        assert pth.startswith(str(test_dir))


def test_wheel_package(copy_test_data):
    td = copy_test_data("package1")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    whl_file = td / "package1-0.0.1-py3-none-any.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)

    with unpack(wf) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        assert files == [
            "__init__.py",
            "data_dir/foo.sh",
            "foo.py",
            "package1-0.0.1.dist-info/METADATA",
            "package1-0.0.1.dist-info/RECORD",
            "package1-0.0.1.dist-info/WHEEL",
            "package1-0.0.1.dist-info/entry_points.txt",
            "package1/__init__.py",
            "package1/data_dir/foo.sh",
            "package1/foo.py",
            "package1/subpkg/__init__.py",
            "package1/subpkg/sp_data_dir/test.json",
            "package1/subpkg2/__init__.py",
            "subpkg/__init__.py",
            "subpkg/sp_data_dir/test.json",
            "subpkg2/__init__.py",
        ]


def test_editable_wheel_package(copy_test_data):
    td = copy_test_data("package1")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td, editable=True)
    whl_file = td / "package1-0.0.1-py3-none-any.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        pth_path = Path(unpacked, "package1.pth")
        check_paths(pth_path, td)
        assert_isdir(Path(unpacked, "package1-0.0.1.dist-info"))


def test_editable_wheel_namespace_package_editable(copy_test_data):
    td = copy_test_data("ns1-pkg")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td, editable=True)
    whl_file = td / "ns1_pkg-1.0-py3-none-any.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        pth_path = Path(unpacked, "ns1.pkg.pth")
        assert_isfile(pth_path)
        lines = pth_path.read_text().splitlines(False)
        assert lines == [str(td / "ns1")]
        assert_isdir(Path(unpacked, "ns1_pkg-1.0.dist-info"))


def test_wheel_module3_with_src_dir(copy_test_data):
    td = copy_test_data("module3")
    wf = make_wheel(
        pyproject_file=td / "pyproject.toml",
        output_dir=td / "dist",
    )
    with unpack(wf) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        assert files == [
            "module3-1.0.dist-info/LICENSE",
            "module3-1.0.dist-info/METADATA",
            "module3-1.0.dist-info/RECORD",
            "module3-1.0.dist-info/WHEEL",
            "src/module3.py",
        ]


def test_editable_wheel_module3_with_src_dir(copy_test_data):
    td = copy_test_data("module3")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td, editable=True)
    whl_file = td / "module3-1.0-py3-none-any.whl"
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        pth_path = Path(unpacked, "module3.pth")
        assert_isfile(pth_path)
        assert pth_path.read_text().strip() == str(td / "src")
        assert_isdir(Path(unpacked, "module3-1.0.dist-info"))


def test_wheel_with_src_package(copy_test_data):
    td = copy_test_data("package2")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    whl_file = td / "package2-1.0.0-py3-none-any.whl"
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        expected = [
            "package2-1.0.0.dist-info/METADATA",
            "package2-1.0.0.dist-info/RECORD",
            "package2-1.0.0.dist-info/WHEEL",
            "package2/__init__.py",
            "package2/foo.py",
        ]
        assert files == expected


def test_editable_wheel_src_package(copy_test_data):
    td = copy_test_data("package2")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td, editable=True)
    whl_file = td / "package2-1.0.0-py3-none-any.whl"
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        pth_path = Path(unpacked, "package2.pth")
        assert_isfile(pth_path)
        assert pth_path.read_text().strip() == str(td / "src")
        assert_isdir(Path(unpacked, "package2-1.0.0.dist-info"))


def test_wheel_ns_package(copy_test_data):
    td = copy_test_data("ns1-pkg")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    assert wf == td / "ns1_pkg-1.0-py3-none-any.whl"
    assert_isfile(wf)
    with unpack(wf) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        assert files == [
            "ns1/pkg/__init__.py",
            "ns1_pkg-1.0.dist-info/METADATA",
            "ns1_pkg-1.0.dist-info/RECORD",
            "ns1_pkg-1.0.dist-info/WHEEL",
        ]


def test_dist_name(copy_test_data):
    td = copy_test_data("altdistname")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    whl_file = td / "packagefoo-1.0-py3-none-any.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)

    with unpack(wf) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        assert files == [
            "package1/__init__.py",
            "package1/data_dir/foo.sh",
            "package1/foo.py",
            "package1/subpkg/__init__.py",
            "package1/subpkg/sp_data_dir/test.json",
            "package1/subpkg2/__init__.py",
            "packagefoo-1.0.dist-info/METADATA",
            "packagefoo-1.0.dist-info/RECORD",
            "packagefoo-1.0.dist-info/WHEEL",
        ]


def test_entry_points(copy_test_data):
    td = copy_test_data("entrypoints_valid")
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    whl_file = td / "package1-1.0-py3-none-any.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)
    with unpack(td / "package1-1.0-py3-none-any.whl") as td_unpack:
        entry_points = Path(td_unpack, "package1-1.0.dist-info", "entry_points.txt")
        assert_isfile(entry_points)
        cp = configparser.ConfigParser()
        cp.read(str(entry_points))
        results = {sec: cp.items(section=sec, raw=True) for sec in cp.sections()}
        expected = {
            "console_scripts": [("pkg_script", "package1:main")],
            "myplugins": [("package1", "package1:main")],
        }
        assert results == expected


def test_wheel_builder():
    # Slightly lower level interface
    with tempfile.TemporaryDirectory() as td:
        wb = WheelBuilder.from_pyproject_file(
            pyproject_file=data_dir / "package1" / "pyproject.toml"
        )
        target = wb.build(output_dir=td)

        assert zipfile.is_zipfile(target)
        assert str(target) == str(Path(td) / "package1-0.0.1-py3-none-any.whl")


@skipIf(os.name == "nt", "Windows does not preserve necessary permissions")
def test_permissions_normed(copy_test_data):
    td = copy_test_data("module1_toml")

    (td / "module1.py").chmod(0o620)
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)

    whl_file = td / "module1-1.0-py3-none-any.whl"
    assert str(wf) == str(whl_file)

    with zipfile.ZipFile(str(whl_file)) as zf:
        info = zf.getinfo("module1.py")
        perms = (info.external_attr >> 16) & 0o777
        assert perms == 0o644, oct(perms)
    whl_file.unlink()

    # This time with executable bit set
    (td / "module1.py").chmod(0o720)
    wf = make_wheel(pyproject_file=td / "pyproject.toml", output_dir=td)
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)
    with zipfile.ZipFile(str(whl_file)) as zf:
        info = zf.getinfo("module1.py")
        perms = (info.external_attr >> 16) & 0o777
        assert perms == 0o755, oct(perms)

        info = zf.getinfo("module1-1.0.dist-info/METADATA")
        perms = (info.external_attr >> 16) & 0o777
        assert perms == 0o644, oct(perms)

        info = zf.getinfo("module1-1.0.dist-info/RECORD")
        perms = (info.external_attr >> 16) & stat.S_IFREG
        assert perms


def test_compression(tmp_path):
    wf = make_wheel(
        pyproject_file=data_dir / "module1_toml" / "pyproject.toml", output_dir=tmp_path
    )
    assert_isfile(wf)
    with zipfile.ZipFile(wf) as zf:
        for name in [
            "module1.py",
            "module1-1.0.dist-info/METADATA",
        ]:
            assert zf.getinfo(name).compress_type == zipfile.ZIP_DEFLATED


def test_wheel_module_local_version_and_custom_tag(copy_test_data):
    """Test if a local version specifier is preserved in wheel filename and dist-info dir name"""
    td = copy_test_data("modulewithlocalversion")
    wf = make_wheel(
        pyproject_file=td / "pyproject.toml",
        output_dir=td,
        wheel_tag="py2.py3-none-manylinux1_x86_64",
    )

    whl_file = td / "modulewithlocalversion-1.0.dev0+test-py2.py3-none-manylinux1_x86_64.whl"
    assert_isfile(whl_file)
    assert str(wf) == str(whl_file)
    with unpack(whl_file) as unpacked:
        assert_isfile(Path(unpacked, "modulewithlocalversion.py"))
        dist_info_dir = Path(unpacked, "modulewithlocalversion-1.0.dev0+test.dist-info")
        assert_isdir(dist_info_dir)
        wheel_tags = (dist_info_dir / "WHEEL").read_text().splitlines(False)
        from flot import __version__

        expected = [
            "Wheel-Version: 1.0",
            f"Generator: flot {__version__}",
            "Root-Is-Purelib: false",
            "Tag: py2.py3-none-manylinux1_x86_64",
        ]
        assert wheel_tags == expected


def test_WheelBuilder_build_with_scripts(
    copy_test_data,
    tmp_path,
):
    td = copy_test_data("scripts")
    builder = WheelBuilder.from_pyproject_file(td / "not-py-project.foo")
    wd = builder.build(tmp_path)

    with unpack(wd) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        assert files == [
            "deep/nested/foo.py",
            "module3-1.0.dist-info/LICENSE",
            "module3-1.0.dist-info/METADATA",
            "module3-1.0.dist-info/RECORD",
            "module3-1.0.dist-info/WHEEL",
            "module3.py",
            "module4.py",
            "somenewfile-wheel.txt",
        ]



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import hashlib
import logging
import os
import re
import sys
from datetime import datetime
from pathlib import Path
from subprocess import check_output

from . import versionno

log = logging.getLogger(__name__)


# For the timestamps, default to 2022-02-02T02:02:02 (UTC)
# This makes the build reproducible. Use the SOURCE_DATE_EPOCH env var with a
# UTC Unix timestamp for an alternative date.

_DT = datetime.fromisoformat("2022-02-02T02:02:02")
FLOT_EPOCH_TIMESTAMP = int(os.environ.get("SOURCE_DATE_EPOCH") or _DT.timestamp())
_DT = datetime.fromtimestamp(FLOT_EPOCH_TIMESTAMP)
# zipfile expects a 6-tuple: neither a Unix timestamp nor datetime object.
FLOT_ZIP_TIME = (
    _DT.year,
    _DT.month,
    _DT.day,
    _DT.hour,
    _DT.minute,
    _DT.second,
)


class FileSelector:
    """
    Compute a set of file Path inclusion or exclusion patterns relative to
    base_dir
    """

    def __init__(
        self,
        base_dir,
        includes,
        excludes=None,
        label=None,
    ):
        self.base_dir = base_dir = Path(base_dir).absolute()

        log.debug(
            f"FileSelector: includes: {includes!r}, "
            f"excludes: {excludes!r}, base_dir: {base_dir!r}"
        )

        if not isinstance(includes, (list, tuple)):
            raise Exception(
                f"Invalid includes patterns: not a list: {type(includes)!r} - {includes!r}"
            )
        self.includes = tuple(includes)

        if excludes:
            if not isinstance(excludes, (list, tuple)):
                raise Exception(
                    f"Invalid excludes patterns: not a list: {type(excludes)!r} - {excludes!r}"
                )
            self.excludes = tuple(excludes)
        else:
            self.excludes = tuple()

        self.label = label or ""

    @property
    def files(self):
        selected_files = set()

        log.info(f"Selecting files for {self.label!r}")

        for pattern in self.includes:
            # Note that a trailing / in Path.glob will return only dirs in
            # Python 3.11 and up
            # pattern = pattern.rstrip("/")
            try:
                selected_files.update(self.base_dir.glob(pattern))
                log.info(
                    f"  Includes pattern: {pattern!r}: " f"{len(selected_files)} files to include"
                )
            except Exception as e:
                raise Exception(f"Invalid pattern: {pattern!r}") from e

        for pattern in self.excludes:
            try:
                selected_files.difference_update(self.base_dir.glob(pattern))
                log.info(
                    f"  Excludes pattern: {pattern!r}: " f"{len(selected_files)} files to exclude"
                )
            except Exception as e:
                raise Exception(f"Invalid pattern: {pattern!r}") from e

        # list of tuples of (absolute Path, relative Path)
        return sorted(
            (path, path.relative_to(self.base_dir))
            for path in selected_files
            if not path.is_dir()
            # TODO: we could also ignore ~ editor/swap files and pyo
            and path.suffix != ".pyc"
        )


class NoVersionError(ValueError):
    pass


class InvalidVersion(ValueError):
    pass


def check_version(version):
    """
    Check whether a given version string match PEP 440, and do normalisation.

    Raise InvalidVersion/NoVersionError with relevant information if
    version is invalid.

    Log a warning if the version is not canonical with respect to PEP 440.

    Returns the version in canonical PEP 440 format.
    """
    if not version:
        raise NoVersionError(
            "Cannot package module without a version string. "
            'Please define a `__version__ = "x.y.z"` in your module.'
        )
    if not isinstance(version, str):
        raise InvalidVersion("__version__ must be a string, not {}.".format(type(version)))

    # Import here to avoid circular import
    version = versionno.normalize_version(version)

    return version


def parse_entry_point(ep):
    """Check and parse a 'package.module:func' style entry point specification.

    Returns (modulename, funcname)
    """
    if ":" not in ep:
        raise ValueError("Invalid entry point (no ':'): %r" % ep)
    mod, func = ep.split(":")

    for piece in func.split("."):
        if not piece.isidentifier():
            raise ValueError("Invalid entry point: %r is not an identifier" % piece)
    for piece in mod.split("."):
        if not piece.isidentifier():
            raise ValueError("Invalid entry point: %r is not a module path" % piece)

    return mod, func


def write_entry_points(d, fp):
    """Write entry_points.txt from a two-level dict

    Sorts on keys to ensure results are reproducible.
    """
    for group_name in sorted(d):
        fp.write("[{}]\n".format(group_name))
        group = d[group_name]
        for name in sorted(group):
            val = group[name]
            fp.write("{}={}\n".format(name, val))
        fp.write("\n")


def hash_file(path, algorithm="sha256"):
    with open(path, "rb") as f:
        h = hashlib.new(algorithm, f.read())
    return h.hexdigest()


def normalize_file_permissions(st_mode):
    """Normalize the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode


class Metadata:
    summary = None
    home_page = None
    author = None
    author_email = None
    maintainer = None
    maintainer_email = None
    license = None
    description = None
    keywords = None
    download_url = None
    requires_python = None
    description_content_type = None

    platform = ()
    supported_platform = ()
    classifiers = ()
    requires = ()
    obsoletes = ()
    project_urls = ()
    provides_dist = ()
    requires_dist = ()
    obsoletes_dist = ()
    requires_external = ()
    provides_extra = ()

    metadata_version = "2.1"

    def __init__(self, data):
        data = data.copy()
        self.name = data.pop("name")
        self.version = data.pop("version")

        for k, v in data.items():
            assert hasattr(self, k), "data does not have attribute '{}'".format(k)
            setattr(self, k, v)

    def _normalise_name(self, n):
        return n.lower().replace("-", "_")

    def write_metadata_file(self, fp):
        """Write out metadata in the email headers format"""
        fields = [
            "Metadata-Version",
            "Name",
            "Version",
        ]
        optional_fields = [
            "Summary",
            "Home-page",
            "License",
            "Keywords",
            "Author",
            "Author-email",
            "Maintainer",
            "Maintainer-email",
            "Requires-Python",
            "Description-Content-Type",
        ]

        for field in fields:
            value = getattr(self, self._normalise_name(field))
            fp.write("{}: {}\n".format(field, value))

        for field in optional_fields:
            value = getattr(self, self._normalise_name(field))
            if value is not None:
                # TODO: verify which fields can be multiline
                # The spec has multiline examples for Author, Maintainer &
                # License (& Description, but we put that in the body)
                # Indent following lines with 8 spaces:
                value = "\n        ".join(value.splitlines())
                fp.write("{}: {}\n".format(field, value))

        for clsfr in self.classifiers:
            fp.write("Classifier: {}\n".format(clsfr))

        for req in self.requires_dist:
            fp.write("Requires-Dist: {}\n".format(req))

        for url in self.project_urls:
            fp.write("Project-URL: {}\n".format(url))

        for extra in self.provides_extra:
            fp.write("Provides-Extra: {}\n".format(extra))

        if self.description is not None:
            fp.write("\n" + self.description + "\n")


def normalize_dist_name(name: str, version: str) -> str:
    """Normalizes a name and a PEP 440 version

    The resulting string is valid as dist-info folder name
    and as first part of a wheel filename

    See https://packaging.python.org/specifications/binary-distribution-format/#escaping-and-unicode
    """
    normalized_name = re.sub(r"[-_.]+", "_", name, flags=re.UNICODE).lower()
    assert check_version(version) == version
    assert "-" not in version, "Normalized versions can’t have dashes"
    return "{}-{}".format(normalized_name, version)


def dist_info_name(distribution, version):
    """Get the correct name of the .dist-info dir"""
    return normalize_dist_name(distribution, version) + ".dist-info"


def run_scripts(pyproject_file, scripts=()):
    """
    Support running arbitrary build scripts at the start of a wheel or sdist
    build.

    The scripts are simple Python scripts that accept a single argument: the
    absolute path to the pyproject.toml file used to run the build.

    They must be available in a location realtive to the pyproject.toml file.

    Scripts may have requirements for extra Python package to use at build time.
    These should be added to the  ``[build-system]`` table requires section.
    """
    for abs_path, _rel_path in scripts:
        script = str(abs_path)
        log.info(f"Running script: {sys.executable} {abs_path} {pyproject_file}")
        check_output([sys.executable, script, pyproject_file])



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import io
import logging
import tarfile
from copy import copy
from gzip import GzipFile
from pathlib import Path
from posixpath import join as pjoin

from . import common

log = logging.getLogger(__name__)


def make_sdist(
    pyproject_file=None,
    output_dir=None,
):
    """
    Make an sdist, return the built archive path.
    """
    builder = SdistBuilder.from_pyproject_file(
        pyproject_file=pyproject_file,
    )
    sdist = builder.build(output_dir=output_dir)
    log.info(f"Built sdist: {sdist}")
    return sdist


class SdistBuilder:
    def __init__(
        self,
        pyproject_file,
        metadata,
        entrypoints,
        includes=(),
        excludes=(),
        metadata_files=(),
        sdist_scripts=(),
        wheel_scripts=(),
        sdist_extra_includes=(),
        sdist_extra_excludes=(),
    ):
        log.debug(f"SdistBuilder: {pyproject_file}")
        pyproject_file = pyproject_file or "pyproject.toml"
        self.pyproject_file = Path(pyproject_file).absolute()
        self.base_dir = self.pyproject_file.parent
        self.metadata = metadata
        self.entrypoints = entrypoints

        self.selected_files = common.FileSelector(
            base_dir=self.base_dir,
            includes=includes,
            excludes=excludes,
            label="base includes/excludes",
        )

        self.selected_metadata_files = common.FileSelector(
            includes=metadata_files,
            base_dir=self.base_dir,
            label="metadata files",
        )

        self.selected_extra_files = common.FileSelector(
            base_dir=self.base_dir,
            includes=sdist_extra_includes,
            excludes=sdist_extra_excludes,
            label="sdist extra",
        )

        self.sdist_scripts = common.FileSelector(
            includes=sdist_scripts,
            base_dir=self.base_dir,
            label="sdist scripts",
        )

        self.wheel_scripts = common.FileSelector(
            includes=wheel_scripts,
            base_dir=self.base_dir,
            label="wheel scripts",
        )

    @classmethod
    def from_pyproject_file(
        cls,
        pyproject_file=None,
    ):
        """
        Create an sdisin target_dir, default: <pyproject file parent>/dist
        from an arbitrary pyproject.toml file path, default: "pyproject.toml".
        """
        from .config import read_pyproject_file

        pyproject_file = pyproject_file or "pyproject.toml"
        pyproject_file = Path(pyproject_file).absolute()
        project_info = read_pyproject_file(pyproject_file)

        return cls(
            pyproject_file=pyproject_file,
            metadata=common.Metadata(project_info.metadata),
            entrypoints=project_info.entrypoints,
            includes=project_info.includes,
            excludes=project_info.excludes,
            metadata_files=project_info.metadata_files,
            sdist_extra_includes=project_info.sdist_extra_includes,
            sdist_extra_excludes=project_info.sdist_extra_excludes,
            sdist_scripts=project_info.sdist_scripts,
            wheel_scripts=project_info.wheel_scripts,
        )

    @property
    def sdist_filename(self):
        return f"{self.dist_name}.tar.gz"

    @property
    def dist_name(self):
        return common.normalize_dist_name(self.metadata.name, self.metadata.version)

    def build(self, output_dir=None):
        """
        Build sdist and return sdist file path.
        """
        common.run_scripts(self.pyproject_file, self.sdist_scripts.files)

        output_dir = (
            output_dir and Path(output_dir).absolute() or Path(self.pyproject_file.parent) / "dist"
        )
        output_dir.mkdir(parents=True, exist_ok=True)
        target = output_dir / self.sdist_filename

        mtime = common.FLOT_EPOCH_TIMESTAMP
        with GzipFile(str(target), mode="wb", mtime=mtime) as gz:
            with tarfile.TarFile(
                str(target),
                mode="w",
                fileobj=gz,
                format=tarfile.PAX_FORMAT,
            ) as tar_file:
                self.add_files(tar_file, mtime)
                self.write_metadata(tar_file, mtime)

        return target

    def add_files(self, tar_file, mtime):
        log.info("Copying selected file(s) to sdist")
        seen_paths = set()
        for i, (abs_path, rel_path) in enumerate(self._select_all_files(), 1):
            if rel_path in seen_paths:
                continue
            self._add_file(
                abs_path=abs_path,
                tar_file=tar_file,
                mtime=mtime,
                rel_path=rel_path,
            )
            seen_paths.add(rel_path)
            if not (i % 500):
                log.info(f"  {i} files added.")

    def _select_all_files(self):
        yield from self.selected_files.files
        yield from self.selected_extra_files.files
        yield from self.selected_metadata_files.files
        yield from self.sdist_scripts.files
        yield from self.wheel_scripts.files
        # Always include the pyproject file at the root renaming to use the
        # standard name
        ppt_abs_path = self.pyproject_file.absolute()
        ppt_rel_path = Path("pyproject.toml")
        yield ppt_abs_path, ppt_rel_path

    def _add_file(self, abs_path, tar_file, mtime, rel_path=None):
        if not rel_path:
            # If not provided, create the rel_path as relative to the pyproject
            # dir otherwise the includes/excludes will no longer work if trying
            # to build a wheel from an sdist
            rel_path = Path(abs_path).relative_to(self.pyproject_dir)

        # log.debug(f"Adding {abs_path!r} to tar file as {rel_path!r}")
        abs_path, rel_path = str(abs_path), str(rel_path)
        ti = tar_file.gettarinfo(abs_path, arcname=pjoin(self.dist_name, rel_path))
        ti = clean_tarinfo(ti, mtime)

        if ti.isreg():
            with open(abs_path, "rb") as f:
                tar_file.addfile(ti, f)
        else:
            tar_file.addfile(ti)  # Symlinks & ?

    def write_metadata(self, tar_file, mtime):
        log.info("Writing PKG-INFO metadata file")

        stream = io.StringIO()
        self.metadata.write_metadata_file(stream)
        pkg_info = stream.getvalue().encode()
        ti = tarfile.TarInfo(pjoin(self.dist_name, "PKG-INFO"))
        ti.size = len(pkg_info)
        ti = clean_tarinfo(ti, mtime)
        tar_file.addfile(ti, io.BytesIO(pkg_info))


def clean_tarinfo(ti, mtime=None):
    """Clean metadata from a TarInfo object to make it more reproducible.

    - Set uid & gid to 0
    - Set uname and gname to ""
    - Normalise permissions to 644 or 755
    - Set mtime if not None
    """
    ti = copy(ti)
    ti.uid = 0
    ti.gid = 0
    ti.uname = ""
    ti.gname = ""
    ti.mode = common.normalize_file_permissions(ti.mode)
    if mtime is not None:
        ti.mtime = mtime
    return ti



import argparse
from pathlib import Path


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("pyproject")
    args = parser.parse_args(argv)
    pyproject = Path(args.pyproject)
    with open(pyproject.parent / "somenewfile-sdist.txt", "w") as out:
        out.write("some content sdist")


if __name__ == "__main__":
    main()



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

if __name__ == "__main__":
    from . import main

    main()



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import os
import os.path as osp
import tarfile
import tempfile
import zipfile
from pathlib import Path

import pytest
from testpath import assert_isdir
from testpath import assert_isfile

from flot import __version__
from flot import buildapi

"""Tests of flot building itself"""


@pytest.fixture()
def cwd_project():
    proj_dir = osp.dirname(osp.dirname(osp.abspath(buildapi.__file__)))
    if not osp.isfile(osp.join(proj_dir, "pyproject.toml")):
        pytest.skip("need flot source directory")

    old_cwd = os.getcwd()
    try:
        os.chdir(proj_dir)
        yield
    finally:
        os.chdir(old_cwd)


def test_prepare_metadata(tmp_path, cwd_project):
    tmp_path = str(tmp_path)
    dist_info = buildapi.prepare_metadata_for_build_wheel(tmp_path)

    assert dist_info.endswith(".dist-info")
    assert dist_info.startswith("flot")
    dist_info = osp.join(tmp_path, dist_info)
    assert_isdir(dist_info)

    assert_isfile(osp.join(dist_info, "WHEEL"))
    assert_isfile(osp.join(dist_info, "METADATA"))


def test_wheel(tmp_path, cwd_project):
    tmp_path = str(tmp_path)
    filename = buildapi.build_wheel(tmp_path)

    assert filename.endswith(".whl")
    assert filename.startswith("flot")
    path = osp.join(tmp_path, filename)
    assert_isfile(path)
    assert zipfile.is_zipfile(path)


def unpack(path):
    z = zipfile.ZipFile(str(path))
    t = tempfile.TemporaryDirectory()
    z.extractall(t.name)
    return t


def test_editable(tmp_path, cwd_project):
    tmp_path = str(tmp_path)
    filename = buildapi.build_editable(tmp_path)

    assert filename.endswith(".whl")
    assert filename.startswith("flot")
    path = osp.join(tmp_path, filename)
    assert_isfile(path)
    assert zipfile.is_zipfile(path)
    with unpack(path) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked)) for p in Path(unpacked).glob("**/*") if p.is_file()
        )
        pth_path = Path(unpacked, "flot.pth")
        assert pth_path.read_text().endswith("flot\n")

        dist_info = f"flot-{__version__}.dist-info"
        assert files == [
            f"{dist_info}/LICENSE",
            f"{dist_info}/METADATA",
            f"{dist_info}/RECORD",
            f"{dist_info}/WHEEL",
            f"{dist_info}/entry_points.txt",
            f"flot.pth",
        ]


def test_sdist(tmp_path, cwd_project):
    tmp_path = str(tmp_path)
    filename = buildapi.build_sdist(tmp_path)

    assert filename.endswith(".tar.gz")
    assert filename.startswith("flot")
    path = osp.join(tmp_path, filename)
    assert_isfile(path)
    assert tarfile.is_tarfile(path)



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import pytest

from flot.common import InvalidVersion
from flot.versionno import normalize_version


def test_normalise_version():
    nv = normalize_version
    assert nv("4.3.1") == "4.3.1"
    assert nv("1.0b2") == "1.0b2"
    assert nv("2!1.3") == "2!1.3"

    # Prereleases
    assert nv("1.0B2") == "1.0b2"
    assert nv("1.0.b2") == "1.0b2"
    assert nv("1.0beta2") == "1.0b2"
    assert nv("1.01beta002") == "1.1b2"
    assert nv("1.0-preview2") == "1.0rc2"
    assert nv("1.0_c") == "1.0rc0"

    # Post releases
    assert nv("1.0post-2") == "1.0.post2"
    assert nv("1.0post") == "1.0.post0"
    assert nv("1.0-rev3") == "1.0.post3"
    assert nv("1.0-2") == "1.0.post2"

    # Development versions
    assert nv("1.0dev-2") == "1.0.dev2"
    assert nv("1.0dev") == "1.0.dev0"
    assert nv("1.0-dev3") == "1.0.dev3"

    assert nv("1.0+ubuntu-01") == "1.0+ubuntu.1"
    assert nv("v1.3-pre2") == "1.3rc2"
    assert nv(" 1.2.5.6\t") == "1.2.5.6"
    assert nv("1.0-alpha3-post02+ubuntu_xenial_5") == "1.0a3.post2+ubuntu.xenial.5"

    with pytest.raises(InvalidVersion):
        nv("3!")

    with pytest.raises(InvalidVersion):
        nv("abc")
